unit Flog;

interface

uses
	SysUtils, Graphics ,Classes, Math
    ;

type
    TFlog = class

    private

        fAngle : double; //  Ă
        fSpeed : double; //  1̈ړňړ鋗
        fTarget: integer;
        fEscape: integer;
        fEscapeAngle: double;

        fColor : TColor;    //  ̐F

        fMagnification: double; //  g嗦

        procedure mDraw();                  //@̕`
        procedure mDrawConnection();        //  ǂ̕`
        procedure mFindTarget();            //  ǂT
        procedure mMove();                  //  1Xebvړ
        procedure mNextState();

    public

        fPosX  : double; //  EɂXW(_E)
        fPosY  : double; //  EɂYW(_)
        fPosZ  : double; //  EɂZW

        Constructor Create();
        Destructor  Destroy(); override;

        procedure mReset(angle: double);    //  

        class function  mGetTotal(): integer;
        class procedure mSetCanvas(pCanvas: TCanvas);
        class procedure mDrawFlogs();
        class procedure mResetAll();
        class procedure mMoveAll();
        class procedure mGetCenter();
    end;

type
    TFlogList = class(TList)
    private
        function Get(Index: Integer): TFlog;
        procedure Put(Index: Integer; const Value: TFlog);
    public
        property Items[Index: Integer]: TFlog read Get write Put; default;
        destructor Destroy; override;
    end;

procedure ResetFlogs();
procedure DrawFlogs();
procedure MoveFlogs();
function  GetTotal(): integer;

var
    fDrawConnection: Boolean = True;

implementation

const
    PI  = System.Pi;            //  180x
    PI2 = PI * 2;               //  360x

    FIT_DISTANCE  = 47.0;       //  H
    MAX_SPEED     = 14.0;       //  ōx

    FLOCK_WEIGHT  = 0.06;       //  H
    CENTER_WEIGHT = 0.007;      //  H
    TARGET_WEIGHT = 0.1;        //  H
    MIMIC_WEIGHT  = 0.1;        //  H

    ACCELERATION  = 1.3;        //  x@H
    VISIBLE_ANGLE = 0.9 * PI;   //@e̂̎̍L  160x

    MAX_FLOG_NUM  = 30;         //  ő̐

    WORLD_WIDTH   = 600;        //  ̕
    WORLD_HEIGHT  = 400;        //  ̍

var
    fFlogList  : TFlogList = nil;  // CX^X̊ǗpXg
    fMarkedFlog: integer;
    fCanvas: TCanvas;

    fCenterX: double;
    fCenterY: double;

(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        TFlogList

    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)

//  ==========================================================
function TFlogList.Get(Index: Integer): TFlog;
begin
    Result :=  TFlog( inherited Get(Index) );
end;

//  ==========================================================
procedure TFlogList.Put(Index: Integer; const Value: TFlog);
begin
    inherited Put( Index, Value );
end;

//  ==========================================================
destructor TFlogList.Destroy;
var
    i: Integer;
begin
    for i := 0 to Self.Count - 1 do
    Self.Items[i].Free;
    inherited Destroy;
end;

(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        TFlog

    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)

//  ==========================================================
Constructor TFlog.Create();
begin
    inherited;
    if fFlogList = nil then begin
        fFlogList := TFlogList.Create;
        fFlogList.Add(Self);
    end else begin
        fFlogList.Add(Self);
    end;
end;

//  ==========================================================
Destructor  TFlog.Destroy;
begin
    fFlogList.Remove(Self);
    if fFlogList.Count = 0 then begin
        fFlogList.Free;
        fFlogList := nil;
    end;
    inherited;
end;

//  ==========================================================
class procedure TFlog.mSetCanvas(pCanvas: TCanvas);
begin
    fCanvas := pCanvas;
end;

function GetTotal: integer;
begin
    Result := 0;
    if fFlogList = nil then Exit;
    Result := fFlogList.Count;  //  CX^X̎擾
end;

class function TFlog.mGetTotal: integer;
begin
    Result := fFlogList.Count;  //  CX^X̎擾
end;

//  ==========================================================
procedure TFlog.mReset(angle: double);
var
    Sin: Extended;
    Cos: Extended;
begin

    SinCos(angle, Sin, Cos);

    fPosX := (0.5 + 0.1 * Cos) * WORLD_WIDTH;
    fPosY := (0.5 + 0.1 * Sin) * WORLD_HEIGHT;

    fAngle := angle + 0.5 * PI;

    fSpeed := 0.3 * MAX_SPEED;

    fTarget := -1;
    fEscape := 0;

end;

class procedure TFlog.mResetAll();
var
    i: integer;
    aFlog: TFlog;
    angle: double;
begin

    if fFlogList = nil then Exit;
    fMarkedFlog := -1;

    for i := 0 to fFlogList.Count - 1 do begin
        aFlog := fFlogList.Items[i];
        angle := PI2 * i / fFlogList.Count;
        aFlog.mReset(angle);
    end;

end;

//  ==========================================================
procedure TFlog.mDraw();

    //  ----------------------------------------
    procedure drawFillCircle(x, y: double; r: integer);
    begin
        fCanvas.Ellipse(
            Round( x - fMagnification * r),          //  S (x, y) Ƃ~
            Round( y - fMagnification * r),
            Round( x + fMagnification {* 2 }* r) + 1,
            Round( y + fMagnification {* 2 }* r) + 1
        );
    end;
    //  ----------------------------------------

var
    xL, xR, vx, yL, yR, vy: double;
begin

    vx := fMagnification * cos(fAngle);
    vy := fMagnification * sin(fAngle);

    xL := fPosX + 28 * vx + 32 * vy;
    yL := fPosY + 28 * vy - 32 * vx;

    xR := fPosX + 28 * vx - 32 * vy;
    yR := fPosY + 28 * vy + 32 * vx;

    with fCanvas do begin
        //  g
        Brush.Color := clBlack;
        drawFillCircle( fPosX, fPosY, 40);
        drawFillCircle( xL, yL, 24);
        drawFillCircle( xR, yR, 24);
        //  {
        Brush.Color := fColor;
        drawFillCircle( fPosX, fPosY, 32);
        drawFillCircle( xL, yL, 16);
        drawFillCircle( xR, yR, 16);
        //  锒
        Brush.Color := clWhite;
        drawFillCircle( xL, yL, 12);
        drawFillCircle( xR, yR, 12);
        //  鍕
        Brush.Color := clBlack;
        drawFillCircle( xL + 8 * vx, yL + 8 * vy, 7);
        drawFillCircle( xR + 8 * vx, yR + 8 * vy, 7);
        //  @
        drawFillCircle( fPosX - 5 * vx + 8 * vy, fPosY - 5 * vy - 8 * vx, 3);
        drawFillCircle( fPosX - 5 * vx - 8 * vy, fPosY - 5 * vy + 8 * vx, 3);
    end;

end;

//  ==========================================================
procedure TFlog.mDrawConnection();
var
    aTarget: TFlog;
begin
    if not fDrawConnection then Exit; 
    if fTarget < 0 then Exit;
    aTarget := fFlogList.Items[fTarget];
    fCanvas.MoveTo(Round(fPosX), Round(fPosY));
    fCanvas.LineTo(Round(aTarget.fPosX), Round(aTarget.fPosY));
end;

//  ==========================================================
class procedure TFlog.mDrawFlogs();
var
    i: integer;
    aFlog: TFlog;
begin
    if fFlogList = nil then Exit;

    for i := 0 to fFlogList.Count - 1 do begin
        aFlog := fFlogList.Items[i];
        if aFlog.fTarget >= 0 then begin
            aFlog.fMagnification := 0.25;
            aFlog.fColor := clLime;
            aFlog.mDraw();
            aFlog.mDrawConnection();
        end else begin
            aFlog.fMagnification := 0.33;
            aFlog.fColor := clRed;
            aFlog.mDraw();
            aFlog.mDrawConnection();
        end;
    end;

end;

function GetAngle(vx, vy: double): double;
var
    a: double;
begin

    //  _ƂāAxɂꍇ
    if (vy = 0) then begin
        if (vx < 0) then begin
            Result := PI;
        end else begin
            Result := 0;
        end;

    //  xł͂Ȃꍇ
    end else begin

        a := ArcCos(vx / sqrt(vx * vx + vy * vy));

        if (vy > 0) then begin
            Result := a;
        end else begin
            Result := PI2 - a;
        end;

    end;

end;

function AngleDifference(t, s: double): double;
var
    a: double;
begin
    a := t - s;
    if a > PI then begin
        Result := a - PI2;
    end else if a < -PI then begin
        Result := a + PI2;
    end else begin
        Result := a;
    end;
end;

procedure TFlog.mFindTarget();
var
    i, target: integer;
    aTarget: TFlog;

    b, d, vx, vy, targetD: double;

begin

    targetD := (0.6 * WORLD_HEIGHT) * (0.6 * WORLD_HEIGHT);
    target  := -1;

    for i := 0 to fFlogList.Count - 1 do begin
        aTarget := fFlogList.Items[i];
        if aTarget = Self then Continue;

        //  ^[Qbg͍Ŋ̑l
        vx := aTarget.fPosX - Self.fPosX;
        vy := aTarget.fPosY - Self.fPosY;
        d  := vx * vx + vy * vy;
        if (d >= targetD) then Continue;

        //  ^[Qbg͕ĂKv
        b := angleDifference(Self.fAngle, aTarget.fAngle);
        if (b <= -0.6 * PI) or (0.6 * PI <= b) then Continue;

        //  ^[Qbg͎EɓĂȂƃ_
        b := AngleDifference(GetAngle(vx, vy), Self.fAngle);
        if (b <= -0.5 * VISIBLE_ANGLE) or (0.5 * VISIBLE_ANGLE <= b) then Continue;

        targetD := d;
        target  := i;
    end;

    Self.fTarget := target;

end;

procedure TFlog.mMove();
begin
    fPosX := fPosX + Cos(fAngle) * fSpeed;
    fPosY := fPosY + Sin(fAngle) * fSpeed;

    if (fPosX < 0) then
        fPosX := fPosX + WORLD_WIDTH
    else if (fPosX > WORLD_WIDTH) then
        fPosX := fPosX - WORLD_WIDTH;

    if (fPosY < 0) then
        fPosY := fPosY + WORLD_HEIGHT
    else if (fPosY > WORLD_HEIGHT) then
        fPosY := fPosY - WORLD_HEIGHT;

end;

class procedure TFlog.mGetCenter();
var
    i: integer;
    x, y: double;
begin
    x := 0;
    y := 0;
    for i := 0 to fFlogList.Count - 1 do begin
        x := x + fFlogList.Items[i].fPosX;
        y := y + fFlogList.Items[i].fPosY;
    end;
    fCenterX := x / fFlogList.Count;
    fCenterY := y / fFlogList.Count;
end;

procedure TFlog.mNextState();
var
    t: integer;
    x, y, a, s, vx, vy: double;
    aTarget: TFlog;
begin
    x := fPosX;
    y := fPosY;
    a := fAngle;
    s := fSpeed;
    t := fTarget;

    if fEscape = 0 then begin
        mFindTarget();
        t := fTarget;

        //  ǂ鑊肪Ȃ
        if t < 0 then begin
            a := a + FLOCK_WEIGHT * AngleDifference(GetAngle(fCenterX - x, fCenterY - y), a);
            s := (1 - 0.7 * Random) * MAX_SPEED;

        //  ǂ鑊肪鎞
        end else begin
            aTarget := fFlogList.Items[fTarget];

            vx := aTarget.fPosX - x;
            vy := aTarget.fPosY - y;

            a := a + CENTER_WEIGHT * AngleDifference(GetAngle(WORLD_WIDTH / 2 - x, WORLD_HEIGHT / 2 - y), a);
            a := a + TARGET_WEIGHT * AngleDifference(GetAngle(vx, vy), a);
            a := a + MIMIC_WEIGHT  * AngleDifference(aTarget.fAngle, a);

            if (vx * vx + vy + vy < FIT_DISTANCE * FIT_DISTANCE) then s := 3 + 7 * Random;

            if (s > MAX_SPEED) then s := MAX_SPEED
            else if (s < 1)    then s := 1;

            if Random < 0.002 then begin
                s := MAX_SPEED;
                if (t = 0) then
                    fEscapeAngle := 0
                else if (AngleDifference(GetAngle(vx, vy), a) < 0) then
                    fEscapeAngle := 0.1 * PI
                else
                    fEscapeAngle := -0.07 * PI;

                t := -1;
                fEscape := 9;

            end;

        end;

    end else begin
        a := a + fEscapeAngle;
        Dec(fEscape);
    end;

    if (a < 0) then
        a := a + PI2
    else if (a >= PI2) then
        a := a - PI2;

    fAngle  := a;
    fSpeed  := s;
    fTarget := t;

end;

class procedure TFlog.mMoveAll();
var
    i: integer;
begin
    for i := 0 to fFlogList.Count - 1 do begin
        fFlogList.Items[i].mMove();
    end;

    mGetCenter;

    for i := 0 to fFlogList.Count - 1 do begin
        fFlogList.Items[i].mNextState();
    end;

end;

procedure ResetFlogs();
begin
    if fFlogList = nil then Exit;
    fFlogList.Items[0].mResetAll;
end;

procedure DrawFlogs();
begin
    if fFlogList = nil then Exit;
    fFlogList.Items[0].mDrawFlogs;
end;

procedure MoveFlogs();
begin
    if fFlogList = nil then Exit;
    fFlogList.Items[0].mMoveAll;
end;

end.
